from .base import TelegramEndpoint
from .commons import DisableNotification


class ForwardMessage(TelegramEndpoint, DisableNotification):
    def __init__(self, chat_id, from_chat_id, message_id):
        args = {
            'chat_id': chat_id,
            'from_chat_id': from_chat_id,
            'message_id': message_id
        }
        super().__init__(args=args)



from .commons import Caption
from ..base import TelegramEndpoint
from ..commons import DisableNotification, ParseMode, ReplyToMessageId, chat_or_inline_message
from ..exceptions import InvalidMediaType
from ..reply_markup import ReplyMarkup
from ...utils import MultipartEncoder


class Send(TelegramEndpoint, DisableNotification, ReplyMarkup, ReplyToMessageId):
    def __init__(self, file, chat_id):
        args = {
            'chat_id': chat_id
        }
        if file.args:
            args.update(file.args)

        endpoint = f'Send{file.__class__.__name__}'

        if file.method == 'POST':
            multipart_encoder = MultipartEncoder()
            multipart_encoder.files = file.value
            boundary, data = multipart_encoder.encode()
            headers = {
                'content-type': f'multipart/form-data; boundary={boundary}'
            }
            super().__init__(method='POST', endpoint=endpoint, args=args, data=data, headers=headers)
        else:
            args[file.file_type] = file.value
            super().__init__(endpoint=endpoint, args=args)


class SendMediaGroup(TelegramEndpoint, DisableNotification, ReplyToMessageId):
    media_types = ['photo', 'video']

    def __init__(self, chat_id, files, json_lib=None):
        if json_lib is None:
            import json
            _json = json
        else:
            _json = json_lib

        self.__multipart_encoder = None
        method = 'GET'
        media = []

        for file in files:
            if file.file_type not in self.media_types:
                raise InvalidMediaType('You can only use SendPhoto and SendVideo in SendMediaGroup')

            serialized, files = file.to_input_media()
            media.append(serialized)
            if files is not None:
                method = 'POST'
                self.multipart_encoder.files.extend(files)

        args = {
            'chat_id': chat_id,
            'media': _json.dumps(media)
        }

        if method == 'POST':
            boundary, data = self.multipart_encoder.encode()
            headers = {
                'content-type': f'multipart/form-data; boundary={boundary}'
            }
            super().__init__(method='POST', args=args, data=data, headers=headers)
        else:
            super().__init__(args=args)

    @property
    def multipart_encoder(self):
        if self.__multipart_encoder is None:
            self.__multipart_encoder = MultipartEncoder()
        return self.__multipart_encoder


class GetFile(TelegramEndpoint):
    # https://api.telegram.org/file/bot<token>/<file_path>
    def __init__(self, file_id):
        args = {
            'file_id': file_id
        }
        super().__init__(args=args)


class EditMessageCaption(TelegramEndpoint, Caption, ParseMode, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id):
        args = chat_or_inline_message(chat_or_inline_message_id, message_id)
        super().__init__(args=args)


class EditMessageMedia(TelegramEndpoint, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id, media):
        args = {
            'media': media.to_input_media(),
            **chat_or_inline_message(chat_or_inline_message_id, message_id)
        }
        super().__init__(args=args)



class CaseInsensitiveDict(dict):
    @staticmethod
    def __lower_key(key):
        return key.lower() if isinstance(key, str) else key

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__convert_keys()

    def __getitem__(self, key):
        return super().__getitem__(self.__lower_key(key))

    def __setitem__(self, key, value):
        super().__setitem__(self.__lower_key(key), value)

    def __delitem__(self, key):
        return super().__delitem__(self.__lower_key(key))

    def __contains__(self, key):
        return super().__contains__(self.__lower_key(key))

    def pop(self, key, *args, **kwargs):
        return super().pop(self.__lower_key(key), *args, **kwargs)

    def get(self, key, *args, **kwargs):
        return super().get(self.__lower_key(key), *args, **kwargs)

    def setdefault(self, key, *args, **kwargs):
        return super().setdefault(self.__lower_key(key), *args, **kwargs)

    def update(self, E=None, **F):
        if E is not None:
            super().update(self.__class__(E))
        if F:
            super().update(self.__class__(**F))

    def __convert_keys(self):
        for k in list(self.keys()):
            v = super().pop(k)
            self.__setitem__(k, v)



from .base import TelegramEndpoint


class SerializableToInlineQueryResult:
    pass


class AnswerInlineQuery(TelegramEndpoint):
    def __init__(self, inline_query_id, results):
        args = {
            'inline_query_id': inline_query_id,
            'results': [x.to_inline_query_result() for x in results]
        }
        super().__init__(args=args)

    def cache_time(self, cache_time):
        self.args['cache_time'] = cache_time
        return self

    def is_personal(self):
        self.args['is_personal'] = True
        return self

    def next_offset(self, next_offset):
        self.args['next_offset'] = next_offset
        return self

    def switch_pm_text(self, switch_pm_text):
        self.args['switch_pm_text'] = switch_pm_text
        return self

    def switch_pm_parameter(self, switch_pm_parameter):
        self.args['switch_pm_parameter'] = switch_pm_parameter
        return self



import asyncio
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from .utils import await_or_call


class Bot:
    update_types = [
        ('message', 'message'),
        ('edited_message', 'message'),
        ('channel_post', 'message'),
        ('edited_channel_post', 'message'),
        ('inline_query', 'inline_query'),
        ('chosen_inline_result', 'chosen_inline_result'),
        ('callback_query', 'callback_query'),
        ('shipping_query', 'shipping_query'),
        ('pre_checkout_query', 'pre_checkout_query')
    ]

    # optional methods
    check_update = None

    def __init__(self, token, max_triggers=1, loop=None):
        """
        The Class(TM).

        :param token: The Telegram-given token
        :param max_triggers: The max number of triggers to execute for a single update
        :param loop: The loop the bot is run into (if it's not asyncio.get_event_loop())
        """

        self.token = token
        self.max_triggers = max_triggers

        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

        self.triggers = []
        self.ignore_check_triggers = []
        self.update_queue = asyncio.PriorityQueue(loop=self.loop)
        self.base_url = 'https://api.telegram.org:443/bot%s/' % token

    async def __api_request(self, method, endpoint, args=None, data=None, headers=None):
        """
        Wraps the urllib3 request in a more user friendly way, making it async and premitting the base Telegram API url.

        :param method: The HTTP method name (GET, POST, PUT, DELETE, HEAD)
        :param endpoint: A Telegram API endpoint (e.g. sendMessage), omitting
            the first slash
        :param args: The get string arguments
        :param data: Binary data to be sent with the request
        :param headers: A CaseInsensitiveDict of headers
        :return: The response from the server
        """

        if args:
            url = f'{self.base_url}{endpoint}?{urlencode(args)}'
        else:
            url = f'{self.base_url}{endpoint}'

        if data and 'content-length' not in headers:
            headers['content-length'] = len(data)

        req = Request(url=url, method=method, data=data, headers=headers)
        res = await self.loop.run_in_executor(None, urlopen, req)
        return res

    async def start(self):
        """
        Main loop

        >>> import asyncio
        >>> loop = asyncio.get_event_loop()
        >>> bot = Bot('token')
        >>> loop.run_until_complete(asyncio.wait([
        >>>     bot.start(),
        >>>     # other tasks
        >>> ]))
        """

        while True:
            _, update = await self.update_queue.get()
            self.loop.create_task(self.handle_update(update))

    async def handle_update(self, update):
        # get type and flavor (if they're none, it's probably an error)
        for _type, flavor in self.update_types:
            if _type in update:
                update_type, update_flavor = _type, flavor
                break
        else:
            update_type, update_flavor = None, None

        # check the update if the function is implemented and skip if it's not passed
        check_update = self.check_update
        if check_update is None:
            checked_triggers = self.triggers + self.ignore_check_triggers
        else:
            is_check_passed = await await_or_call(check_update, update, update_type, update_flavor)
            if is_check_passed is True:
                checked_triggers = self.triggers + self.ignore_check_triggers
            else:
                checked_triggers = self.triggers

        for trigger in checked_triggers:

            is_update_matched = await await_or_call(trigger.match, update, update_type, update_flavor)

            if is_update_matched is not True:
                continue

            # if a matching one is found call the corresponding handler
            endpoint = await await_or_call(trigger.handle, update, update_type, update_flavor)

            if endpoint is None:
                continue

            method, url, args, data, headers = endpoint

            response = await self.__api_request(method, url, args, data, headers)

            # call the callback if existing
            callback = trigger.callback

            if callback is not None:
                await await_or_call(callback, response)

            break

    def push_update(self, update):
        """
        Pushes an update (already json decoded) into the queue.

        :param update: The update to be pushed in the queue
        """

        self.update_queue.put_nowait((update['update_id'], update))

    def trigger(self, ignore_check=False):
        """
        Decorates a Trigger, inserting it into the bot check list

        :param ignore_check: ignores a possible "check_update" function
        """

        def decorator(trigger):
            if ignore_check is True:
                self.ignore_check_triggers.append(trigger())
            else:
                self.triggers.append(trigger())

            return trigger

        return decorator



from .case_insensitive_dict import CaseInsensitiveDict  # noqa: F401
from .commons import await_or_call, not_implemented, random_string  # noqa: F401
from .multipart_encoder import MultipartEncoder  # noqa: F401



from .base import TelegramEndpoint
from .commons import chat_or_inline_message
from .exceptions import InvalidInlineKeyboardButtonPosition, InvalidInlineKeyboardButtonType


class InlineKeyboard:
    button_types = ['url', 'callback_data', 'switch_inline_query', 'switch_inline_query_current_chat', 'callback_game',
                    'pay']
    must_be_first_button_types = ['callback_game', 'pay']

    def __init__(self):
        self.buttons = []

    def serialize(self):
        return {'inline_keyboard': self.buttons}

    def add_button(self, text, button_type, button_value):
        if button_type not in self.button_types:
            raise InvalidInlineKeyboardButtonType

        if button_type in self.must_be_first_button_types and self.buttons:
            raise InvalidInlineKeyboardButtonPosition('You must insert a callback_game/pay button as the first button')

        self.buttons.append({
            'text': text,
            button_type: button_value
        })
        return self


class ReplyKeyboard:
    def __init__(self):
        self.buttons = []
        self.options = {}

    def serialize(self):
        return {
            'keyboard': self.buttons,
            **self.options
        }

    def resize_keyboard(self):
        self.options['resize_keyboard'] = True
        return self

    def one_time_keyboard(self):
        self.options['one_time_keyboard'] = True
        return self

    def selective(self):
        self.options['selective'] = True
        return self

    def add_button(self, text, request_contact=False, request_location=False):
        button = {
            'text': text
        }

        if request_contact is True:
            button['request_contact'] = True

        if request_location is True:
            button['request_location'] = True

        self.buttons.append(button)
        return self


class ReplyKeyboardRemove:
    def __init__(self):
        self.options = {}

    def serialize(self):
        return {
            'remove_keyboard': True,
            **self.options
        }

    def selective(self):
        self.options['selective'] = True
        return self


class ForceReply:
    def __init__(self):
        self.options = {}

    def serialize(self):
        return {
            'force_reply': True,
            **self.options
        }

    def selective(self):
        self.options['selective'] = True
        return self


class ReplyMarkup:
    def reply_markup(self, reply_markup):
        self.args['reply_markup'] = reply_markup.serialize()
        return self


class AnswerCallbackQuery(TelegramEndpoint):
    def __init__(self, callback_query_id):
        args = {
            'callback_query_id': callback_query_id
        }
        super().__init__(args=args)

    def text(self, text):
        self.args['text'] = text
        return self

    def show_alert(self):
        self.args['show_alert'] = True
        return self

    def url(self, url):
        self.args['url'] = url
        return self

    def cache_time(self, cache_time):
        self.args['cache_time'] = cache_time
        return self


class EditMessageReplyMarkup(TelegramEndpoint, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id):
        super().__init__(args=chat_or_inline_message(chat_or_inline_message_id, message_id))



from .base import TelegramEndpoint
from .commons import DisableNotification, ReplyToMessageId
from .reply_markup import ReplyMarkup


class SendContact(TelegramEndpoint, DisableNotification, ReplyMarkup, ReplyToMessageId):
    def __init__(self, chat_id, phone_number, first_name):
        args = {
            'chat_id': chat_id,
            'phone_number': phone_number,
            'first_name': first_name  # noqa: S001
        }
        super().__init__(args=args)

    def last_name(self, last_name):
        self.args['last_name'] = last_name
        return self

    def vcard(self, vcard):
        self.args['vcard'] = vcard
        return self



from .base import TelegramEndpoint
from .exceptions import InvalidChatAction


class SendChatAction(TelegramEndpoint):
    actions = ['typing', 'upload_photo', 'record_video', 'upload_video', 'record_audio', 'upload_audio',
               'upload_document', 'find_location', 'record_video_note', 'upload_video_note']

    def __init__(self, chat_id, action):
        if action not in self.actions:
            raise InvalidChatAction
        args = {
            'chat_id': chat_id,
            'action': action  # noqa: S001
        }
        super().__init__(args=args)



import asyncio
import random
import string


async def await_or_call(func, *args, **kwargs):
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    else:
        return func(*args, **kwargs)


def not_implemented(func):
    def decorated():
        raise NotImplementedError
    return decorated


def random_string(size=16):
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(size))



from .base import TelegramEndpoint
from .commons import DisableNotification
from ..utils import MultipartEncoder


class ExportChatInviteLink(TelegramEndpoint):
    def __init__(self, chat_id):
        args = {
            'chat_id': chat_id
        }
        super().__init__(args=args)


class SetChatPhoto(TelegramEndpoint):
    def __init__(self, chat_id, photo):
        args = {
            'chat_id': chat_id
        }
        self.multipart_encoder = MultipartEncoder()
        self.multipart_encoder.add_file('photo', photo)
        boundary, data = self.multipart_encoder.encode()
        headers = {
            'content-type': f'multipart/form-data; boundary={boundary}'
        }
        super().__init__(method='POST', args=args, data=data, headers=headers)


class DeleteChatPhoto(TelegramEndpoint):
    def __init__(self, chat_id):
        args = {
            'chat_id': chat_id
        }
        super().__init__(args=args)


class SetChatTitle(TelegramEndpoint):
    def __init__(self, chat_id, title):
        args = {
            'chat_id': chat_id,
            'title': title
        }
        super().__init__(args=args)


class SetChatDescription(TelegramEndpoint):
    def __init__(self, chat_id, description):
        args = {
            'chat_id': chat_id,
            'description': description
        }
        super().__init__(args=args)


class PinChatMessage(TelegramEndpoint, DisableNotification):
    def __init__(self, chat_id, message_id):
        args = {
            'chat_id': chat_id,
            'message_id': message_id
        }
        super().__init__(args=args)


class UnpinChatMessage(TelegramEndpoint):
    def __init__(self, chat_id):
        args = {
            'chat_id': chat_id
        }
        super().__init__(args=args)


class LeaveChat(TelegramEndpoint):
    def __init__(self, chat_id):
        args = {
            'chat_id': chat_id
        }
        super().__init__(args=args)


class GetChat(TelegramEndpoint):
    def __init__(self, chat_id):
        args = {
            'chat_id': chat_id
        }
        super().__init__(args=args)


class GetChatAdministrators(TelegramEndpoint):
    def __init__(self, chat_id):
        args = {
            'chat_id': chat_id
        }
        super().__init__(args=args)


class GetChatMemberCount(TelegramEndpoint):
    def __init__(self, chat_id):
        args = {
            'chat_id': chat_id
        }
        super().__init__(args=args)


class GetChatMember(TelegramEndpoint):
    def __init__(self, chat_id, user_id):
        args = {
            'chat_id': chat_id,
            'user_id': user_id
        }
        super().__init__(args=args)


class SetChatStickerSet(TelegramEndpoint):
    def __init__(self, chat_id, sticker_set_name):
        args = {
            'chat_id': chat_id,
            'sticker_set_name': sticker_set_name
        }
        super().__init__(args=args)


class DeleteChatStickerSet(TelegramEndpoint):
    def __init__(self, chat_id):
        args = {
            'chat_id': chat_id
        }
        super().__init__(args=args)



from abc import ABC

from ..utils import CaseInsensitiveDict


class TelegramEndpoint(ABC):
    def __init__(self, method: str = 'GET', endpoint: str = None, args: dict = None, data: bytes = None,
                 headers: dict = None, **kwargs):
        if endpoint is None:
            _classname = self.__class__.__name__
            self.endpoint = f'{_classname[0].lower()}{_classname[1:]}'
        else:
            self.endpoint = endpoint

        self.method = method
        self.args = args if args is not None else {}
        self.data = data
        self.headers = CaseInsensitiveDict(headers) if headers is not None else CaseInsensitiveDict()

        if kwargs:
            for key, value in kwargs.items():
                method = getattr(self, key)
                method(value)

    def serialize(self):
        return self.method, self.endpoint, self.args, self.data, self.headers



from .chat import ExportChatInviteLink, SetChatPhoto, DeleteChatPhoto, SetChatTitle, SetChatDescription, PinChatMessage, UnpinChatMessage, LeaveChat, GetChat, GetChatAdministrators, GetChatMemberCount, GetChatMember, SetChatStickerSet, DeleteChatStickerSet  # noqa: F401, I101, E501
from .chat_action import SendChatAction  # noqa: F401, I101, E501
from .chat_member import KickChatMember, UnbanChatMember, RestrictChatMember, PromoteChatMember  # noqa: F401, I101, E501
from .contact import SendContact  # noqa: F401, I101, E501
from .file import *  # noqa: F401, I101, F403
from .forward import ForwardMessage  # noqa: F401, I101, E501
from .inline_mode import AnswerInlineQuery  # noqa: F401, I101, E501
from .location import SendLocation, EditMessageLiveLocation, StopMessageLiveLocation  # noqa: F401, I101, E501
from .reply_markup import InlineKeyboard, ReplyKeyboard, ReplyKeyboardRemove, ForceReply, AnswerCallbackQuery  # noqa: F401, I101, E501
from .text import SendMessage, EditMessageText  # noqa: F401, I101, E501
from .uncategorized import GetMe, GetUserProfilePhotos, DeleteMessage  # noqa: F401, I101, E501
from .venue import SendVenue  # noqa: F401, I101, E501
from .webhook import SetWebhook, DeleteWebhook, GetWebhookInfo  # noqa: F401, I101, E501

from .exceptions import *  # noqa: F401, I202, F403, I100



from .endpoints import Send, SendMediaGroup, GetFile, EditMessageCaption, EditMessageMedia  # noqa: F401, I101
from .types import Photo, Audio, Document, Video, Animation, Voice, VideoNote  # noqa: F401, I101



from abc import ABC

from .commons import Caption
from ..commons import ParseMode
from ...utils import MultipartEncoder, random_string


class File(ABC):
    file_type = None
    attributes = []
    cache = {}

    def __init__(self, file, is_path=True, **kwargs):
        # cache is a dict {file_path: file_id} and should be updated with the trigger callback
        self.args = {}
        self.file = file

        if is_path is True:
            file_id = self.cache.get(self.file, None)
            if file_id is not None:
                self.method = 'GET'
                self.value = file_id
            else:
                self.method = 'POST'
                self.calls_before_get = 1
                self.value = [MultipartEncoder.encode_file(self.file_type, self.file)]
        else:
            self.method = 'GET'
            self.value = self.file

        if kwargs:
            for key, value in kwargs.items():
                method = getattr(self, key)
                method(value)

    def with_cache(self):
        if self.calls_before_get == -1:
            return self

        if self.calls_before_get != 0:
            self.calls_before_get = 0
            return self

        file_id = self.cache.get(self.file, None)
        self.calls_before_get = -1

        if file_id is not None:
            self.method = 'GET'
            self.value = file_id

        return self

    @staticmethod
    def _get_file_id(response):
        return response

    def update_cache(self, response, index=None):
        file_id = self.cache.get(self.file, None)
        if file_id is None:
            _res = response['result']
            if index is not None:
                _res = _res[index][self.file_type]
            else:
                _res = _res[self.file_type]

            self.cache[self.file] = self._get_file_id(_res)['file_id']


class ToInputMedia:
    media_types = ['photo', 'video']

    def to_input_media(self):

        serialized = {
            'type': self.file_type,
            **self.args
        }

        if self.method == 'POST':
            files = []
            for file in self.value:
                field_name = random_string()
                if file[0] in self.media_types:
                    serialized['media'] = f'attach://{field_name}'
                else:
                    serialized[file[0]] = f'attach://{field_name}'
                files.append((field_name, file[1], file[2], file[3]))
            return serialized, files
        else:
            serialized['media'] = self.value
            return serialized, None


class Thumb:
    def thumb(self, thumb):
        # thumb must be a path
        if self.method == 'POST':
            self.value.append(MultipartEncoder.encode_file('thumb', thumb))
        return self


class Photo(File, Caption, ParseMode, ToInputMedia):
    file_type = 'photo'
    serialized_fields = ['caption', 'parse_mode']

    @staticmethod
    def _get_file_id(response):
        return response[-1]


class Audio(File, Caption, ParseMode, Thumb, ToInputMedia):
    file_type = 'audio'
    attributes = ['duration', 'performer', 'title']
    serialized_fields = ['caption', 'parse_mode', 'thumb']


class Document(File, Caption, ParseMode, Thumb, ToInputMedia):
    file_type = 'document'
    serialized_fields = ['caption', 'parse_mode', 'thumb']


class Video(File, Caption, ParseMode, Thumb, ToInputMedia):
    file_type = 'video'
    attributes = ['duration', 'width', 'height']
    serialized_fields = ['caption', 'parse_mode', 'thumb', 'supports_streaming']

    def supports_streaming(self):
        self.args['supports_streaming'] = True
        return self


class Animation(File, Caption, ParseMode, Thumb, ToInputMedia):
    file_type = 'animation'
    attributes = ['duration', 'width', 'height']
    serialized_fields = ['caption', 'parse_mode', 'thumb']


class Voice(File, Caption, ParseMode):
    file_type = 'voice'
    attributes = ['duration']


class VideoNote(File, Thumb):
    file_type = 'video_note'
    attributes = ['duration', 'length']



from .base import TelegramEndpoint


class GetMe(TelegramEndpoint):
    pass


class GetUserProfilePhotos(TelegramEndpoint):
    def __init__(self, user_id):
        args = {
            'user_id': user_id
        }
        super().__init__(args=args)

    def offset(self, offset):
        self.args['offset'] = offset
        return self

    def limit(self, limit):
        self.args['limit'] = limit
        return self


class DeleteMessage(TelegramEndpoint):
    def __init__(self, chat_id, message_id):
        args = {
            'chat_id': chat_id,
            'message_id': message_id
        }
        super().__init__(args=args)



class InvalidInlineKeyboardButtonType(Exception):
    pass


class InvalidInlineKeyboardButtonPosition(Exception):
    pass


class InvalidChatAction(Exception):
    pass


class InvalidParseMode(Exception):
    pass


class InvalidUpdateType(Exception):
    pass


class InvalidMediaType(Exception):
    pass



from .exceptions import InvalidParseMode


def chat_or_inline_message(chat_or_inline_message_id, message_id):
    if message_id is not None:
        return {
            'chat_id': chat_or_inline_message_id,
            'message_id': message_id
        }
    else:
        return {
            'inline_message_id': chat_or_inline_message_id
        }


class ReplyToMessageId:
    def reply_to_message_id(self, message_id):
        self.args['reply_to_message_id'] = message_id
        return self


class DisableNotification:
    def disable_notification(self):
        self.args['disable_notification'] = True
        return self


class ParseMode:
    parse_modes = ['Markdown', 'HTML']

    def parse_mode(self, parse_mode):
        if parse_mode not in self.parse_modes:
            raise InvalidParseMode

        self.args['parse_mode'] = parse_mode
        return self
